/*
 * Copyright (C) 2010 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 * 
 * This file contains adapted source code from the AOSP project
 * modified by Adam Skillen to provide the password recovery functionality.
 *   
 * Copyright (C) 2013 Adam Skillen
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Adam Skillen
 * 4. Neither the name of Concordia University, Mobiflage, nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *  
 */
 
/*
 * FILE:		cryptfsmount.c
 * 
 * AUTHOR:		Adam Skillen
 * DATE:		17-Feb-12
 *
 * DESCRIPTION:
 * Decrypts and mounts an Android encrypted filesystem image to the given
 * mount-point.  Takes as input the password (which may be a PIN, passphrase,
 * or master volume key derived from PBKDF2), an Android image file, a mount point, and 
 * the type of unlock code (i.e., pin or key).  The filesystem image is loaded RDONLY
 * as a loopback block device, and a dm-crypt devmapper device is created to perform
 * the On-The-Fly-Encryption.
 *
 *  USAGE:		cryptfsmount  <passwd>  <path_to_image>  <mount_point>  <pwd|key>
 *  BUILD:		gcc -o cryptfsmount cryptfsmount.c -lcrypto
 * 
 * NOTICE: 
 * Some of this code is derived from the Android Open Source Project (AOSP).
 * Please see the [aosp_src]/system/vold directory for the original files.
 * The AOSP license information can be found above this text.
*/


#define _LARGEFILE64_SOURCE 1

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <linux/types.h>
#include <linux/dm-ioctl.h>
#include <linux/loop.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <fcntl.h>
#include <unistd.h>
#include <openssl/evp.h>
#include <openssl/hmac.h>


#define HASH_COUNT 2000
#define KEY_LEN_BYTES 16
#define IV_LEN_BYTES 16
#define CRYPT_FOOTER_OFFSET 0x4000
#define SALT_LEN 16
#define KEY_TO_SALT_PADDING 32
#define PROPERTY_VALUE_MAX  92
#define CRYPT_MNT_MAGIC 0xD0B5B1C4
#define DM_CRYPT_BUF_SIZE 4096
#define MAXPATHLEN 64
#define DEVMAPPER_BUFFER_SIZE 4096

#define __le32 unsigned int
#define __le16 unsigned short int 

// crypto footer structure format
struct crypt_mnt_ftr {
  __le32 magic;		/* crypto footer magic number */
  __le16 major_version; 
  __le16 minor_version;
  __le32 ftr_size; 	/* footer size in bytes (not including key) */
  __le32 flags;		/* flags for Android OS */
  __le32 keysize;	/* keylength in bytes */
  __le32 spare1;	/* ignored */
  __le64 fs_size;	/* Size of the encrypted fs, in 512 byte sectors */
  __le32 failed_decrypt_count;
  unsigned char crypto_type_name[64];	/* cipher, mode, IV method, etc. */
};



/* Convert a binary string of specified length into an ascii hex string equivalent */
void bin_to_hex(unsigned char *bin_buf, unsigned int length,
                              char *hex_ascii)
{
  unsigned int i, a;
  unsigned char nibble;

  for (i=0, a=0; i<length; i++, a+=2) {
    /* For each byte, write out two ascii hex digits */
    nibble = (bin_buf[i] >> 4) & 0xf;
    hex_ascii[a] = nibble + (nibble > 9 ? 0x37 : 0x30);

    nibble = bin_buf[i] & 0xf;
    hex_ascii[a+1] = nibble + (nibble > 9 ? 0x37 : 0x30);
  }
  hex_ascii[a] = '\0';
}


/* Turn the password into a key and IV that can decrypt the master key 
   Using OpenSSL PBKDF2 library function given a password and salt and
   iterating 2000 times */
static void pbkdf2(char *passwd, unsigned char *salt, unsigned char *ikey)
{
    PKCS5_PBKDF2_HMAC_SHA1(passwd, strlen(passwd), salt, SALT_LEN,
                           2000, KEY_LEN_BYTES+IV_LEN_BYTES, ikey);
}


/* Read the footer from a disk image and store it in a footer struct */
static int read_crypt_ftr(char *disk_image, struct crypt_mnt_ftr *crypt_ftr,
                                  unsigned char *key, unsigned char *salt)
{
  int fd;
  unsigned int cnt;
  off64_t off;
  int rc = -1;
  struct stat64 statbuf;

  // open the disk image for reading
  if ( (fd = open(disk_image, O_RDONLY)) < 0) {
    fprintf(stderr, "ERROR: Cannot open disk image %s\n", disk_image);
    return -1;
  }

  // get the stat info for the image
  if (fstat64(fd, &statbuf) < 0) {
    fprintf(stderr, "ERROR: Cannot stat(2) disk image %s\n", disk_image);
    goto errout;
  }

  off = ((off64_t)statbuf.st_size) - CRYPT_FOOTER_OFFSET;

  // seek to the footer area in the image
  if (lseek64(fd, off, SEEK_SET) == -1) {
    fprintf(stderr, "ERROR: Cannot seek to footer\n");
    goto errout;
  }

  // read the footer and store it in a struct
  if ( (cnt = read(fd, crypt_ftr, sizeof(struct crypt_mnt_ftr))) != sizeof(struct crypt_mnt_ftr)) {
    fprintf(stderr, "ERROR: Cannot read disk image %s\n", disk_image);
    goto errout;
  }

  // sanity check: ensure proper magic number in footer
  if (crypt_ftr->magic != CRYPT_MNT_MAGIC) {
    fprintf(stderr, "ERROR: Bad magic number in crypt footer: \n Wanted %08x \n Found %08x\n", CRYPT_MNT_MAGIC, crypt_ftr->magic);
    goto errout;
  }

  // if the footer has custom length, skip to end of the footer to read the key and salt
  if (crypt_ftr->ftr_size > sizeof(struct crypt_mnt_ftr)) {
    if (lseek(fd, crypt_ftr->ftr_size - sizeof(struct crypt_mnt_ftr),  SEEK_CUR) == -1) {
      fprintf(stderr, "ERROR: Cannot seek to key\n");
      goto errout;
    }
  }

  // sanity check: ensure proper key length is used (causes a warning, does not halt)
  if (crypt_ftr->keysize != KEY_LEN_BYTES) {
    fprintf(stderr, "WARNING: Keysize of %d bits may not be supported for image %s\n",
          crypt_ftr->keysize * 8, disk_image);
  }

  // read the key stored after the footer
  if ( (cnt = read(fd, key, crypt_ftr->keysize)) != crypt_ftr->keysize) {
    fprintf(stderr, "ERROR: Cannot read key from image %s\n", disk_image);
    goto errout;
  }

  // seek to salt stored after the key
  if (lseek64(fd, KEY_TO_SALT_PADDING, SEEK_CUR) == -1) {
    fprintf(stderr, "ERROR: Cannot seek to salt\n");
    goto errout;
  }

  // read the salt (used for PBKDF2)
  if ( (cnt = read(fd, salt, SALT_LEN)) != SALT_LEN) {
    fprintf(stderr, "ERROR: Cannot read salt from image %s\n", disk_image);
    goto errout;
  }
  rc = 0;

errout:
  close(fd);
  return rc;
}


/* using the PBKDF2 derived key and salt, decrypt the master key */
static int decrypt_master_key(unsigned char *ikey,
                              unsigned char *encrypted_master_key,
                              unsigned char *decrypted_master_key)
{
  EVP_CIPHER_CTX d_ctx;
  int decrypted_len, final_len;

  /* Initialize the decryption engine */
  if (! EVP_DecryptInit(&d_ctx, EVP_aes_128_cbc(), ikey, ikey+KEY_LEN_BYTES)) {
    return -1;
  }
  EVP_CIPHER_CTX_set_padding(&d_ctx, 0); /* Turn off padding as our data is block aligned */

  /* Decrypt the master key */
  if (! EVP_DecryptUpdate(&d_ctx, decrypted_master_key, &decrypted_len,
                            encrypted_master_key, KEY_LEN_BYTES)) {
    return -1;
  }
  if (! EVP_DecryptFinal(&d_ctx, decrypted_master_key + decrypted_len, &final_len)) {
    return -1;
  }

  // sanity check: is the decrypted string the proper length
  if (decrypted_len + final_len != KEY_LEN_BYTES) {
    return -1;
  } else {
    return 0;
  }
}


/* Initialize a dm_ioctl structure for use with dm-crypt devices and the ioctl(2) function */
static void ioctl_init(struct dm_ioctl *io, size_t dataSize, const char *name, unsigned flags)
{
    memset(io, 0, dataSize);
    io->data_size = dataSize;
    io->data_start = sizeof(struct dm_ioctl);
    io->version[0] = 4;
    io->version[1] = 0;
    io->version[2] = 0;
    io->flags = flags;
    if (name) {
        strncpy(io->name, name, sizeof(io->name));
    }
}


/* Using the image loopback device and decrypted master key, create a device mapper
   device to perform On-The-Fly-Encryption functionality */
static int mkdevmap(struct crypt_mnt_ftr *crypt_ftr, unsigned char *master_key,
                                    char *real_blk_name, char *crypto_blk_name, const char *name)
{
  char buffer[DM_CRYPT_BUF_SIZE];
  char master_key_ascii[129]; /* Large enough to hold 512 bit key and null */
  char *crypt_params;
  struct dm_ioctl *io;
  struct dm_target_spec *tgt;
  unsigned int minor;
  int fd;
  int retval = -1;

  // open the devmapper control device to create a new dm-crypt device
  if ((fd = open("/dev/mapper/control", O_RDWR)) < 0 ) {
    fprintf(stderr, "ERROR: Cannot open device-mapper\n");
    goto errout;
  }

  io = (struct dm_ioctl *) buffer;

  // initialize the ioctl struct used for manipulating the dm-crypt device
  ioctl_init(io, DM_CRYPT_BUF_SIZE, name, 0);

  // create a new dm-crypt device
  if (ioctl(fd, DM_DEV_CREATE, io)) {
	fprintf(stderr, "ERROR: Cannot create dm-crypt device (%d)\n", errno);
    goto errout;
  }

  // Get the name of the dm-crypt device file (referenced as /dev/mapper/<name>)
  ioctl_init(io, DM_CRYPT_BUF_SIZE, name, 0);
  if (ioctl(fd, DM_DEV_STATUS, io)) {
	fprintf(stderr, "ERROR: Cannot retrieve dm-crypt device status\n");
    goto errout;
  }
  
  // get dm-crypt block device minor number (referenced as /dev/dm-<minor>)
  minor = (io->dev & 0xff) | ((io->dev >> 12) & 0xfff00);
  snprintf(crypto_blk_name, MAXPATHLEN, "/dev/dm-%u", minor);

  /* Create the mapping table that will be used for this device
     This includes the cipher spec, crypto parameters, key, etc. 
     dm-crypt will require a dm_ioctl and dm_target structures
     we are storing them in a buffer at appropriate offsets    */
  tgt = (struct dm_target_spec *) &buffer[sizeof(struct dm_ioctl)];

  ioctl_init(io, 4096, name, 0);
  io->target_count = 1;
  tgt->status = 0;
  tgt->sector_start = 0;
  tgt->length = crypt_ftr->fs_size;
  strcpy(tgt->target_type, "crypt");

  // move pointer to end of dm_ioctl and dm_target structures in buffer
  crypt_params = buffer + sizeof(struct dm_ioctl) + sizeof(struct dm_target_spec);
  
  // covert binary key to ASCII
  bin_to_hex(master_key, crypt_ftr->keysize, master_key_ascii);     

  // add cryto parameters to buffer
  sprintf(crypt_params, "%s %s 0 %s 0", crypt_ftr->crypto_type_name,
          master_key_ascii, real_blk_name);

  // move pointer to end of crypto parameters
  crypt_params += strlen(crypt_params) + 1;

  /* Align to an 8 byte boundary */
  crypt_params = (char *) (((unsigned long)crypt_params + 7) & ~8); 

  // move tgt.next pointer to end of crypt params
  tgt->next = crypt_params - buffer;

  // Load the mapping table into the devmapper device
  if (ioctl(fd, DM_TABLE_LOAD, io)) {
	fprintf(stderr, "ERROR: Cannot load dm-crypt mapping table.\n");
      goto errout;
  }

  /* Resume the dm-crypt device to activate it */
  ioctl_init(io, 4096, name, 0);
  if (ioctl(fd, DM_DEV_SUSPEND, io)) {
	fprintf(stderr, "ERROR: Cannot resume the dm-crypt device\n");
    goto errout;
  }

  retval = 0;

errout:
  close(fd);

  return retval;
}


/* Destroy a dm-crypt device given it's name (i.e., /dev/mapper/<name> */
int rmdevmap(const char *name) {
    char *buffer = (char *) malloc(DEVMAPPER_BUFFER_SIZE);
    int fd, i;
    struct dm_ioctl *io = (struct dm_ioctl *) buffer;

    // malloc failed
    if (!buffer) {
        fprintf(stderr, "Error allocating memory (%s)", strerror(errno));
        return -1;
    }

    // open the devmapper control device to manipulate the dm-crypt device
    if ((fd = open("/dev/mapper/control", O_RDWR)) < 0) {
        fprintf(stderr, "Error opening devmapper (%s)", strerror(errno));
        free(buffer);
        return -1;
    }

    // Create the dm_ioctl structure for the device
    ioctl_init(io, DEVMAPPER_BUFFER_SIZE, name, 0);

    // Use ioctl to destroy the dm-crypt device
    if (ioctl(fd, DM_DEV_REMOVE, io)) {
        if (errno != ENXIO) {
            fprintf(stderr, "Error destroying device mapping (%s)", strerror(errno));
        }
        free(buffer);
        close(fd);
        return -1;
    }

    free(buffer);
    close(fd);
    return 0;
}


// Create a loopback device for use with the Android filesystem image file
int mkloop(const char *loopFile) {
    int i, loop_fd, image_fd;
    char filename[256];
    mode_t mode = 0660 | S_IFBLK;
    struct loop_info64 li;

    /* using static device name loop254 since we only need one loopback
       and 254 is likely not in use */
    sprintf(filename, "/dev/loop254");

    // create loopback block device
    if (mknod(filename, mode, 254) < 0) {
        if (errno != EEXIST) {
          fprintf(stderr, "Error creating loop device node (%s)", strerror(errno));
          return -1;
        }
    }

    // open the created loopback device
    if ((loop_fd = open(filename, O_RDWR)) < 0) {
        fprintf(stderr, "ERROR: Unable to open %s (%s)", filename, strerror(errno));
        return -1;
    }

    // open the disk image file
    if ((image_fd = open(loopFile, O_RDWR)) < 0) {
        fprintf(stderr, "ERROR: Unable to open %s (%s)", loopFile, strerror(errno));
        close(loop_fd);
        return -1;
    }

    // map the image file to the loopback device file
    if (ioctl(loop_fd, LOOP_SET_FD, image_fd) < 0) {
        fprintf(stderr, "Error setting up loopback interface (%s)", strerror(errno));
        close(image_fd);
        close(loop_fd);
        return -1;
    }

    // initialize a loopback info struct and set some properties
    memset(&li, 0, sizeof(li));
    strncpy((char*) li.lo_crypt_name, "test", LO_NAME_SIZE);
    strncpy((char*) li.lo_file_name, loopFile, LO_NAME_SIZE);

    // configure loopback device to use info struct
    if (ioctl(loop_fd, LOOP_SET_STATUS64, &li) < 0) {
        fprintf(stderr, "Error setting loopback status (%s)", strerror(errno));
        close(image_fd);
        close(loop_fd);
        return -1;
    }

    // sanity check: ensure the loopback configuration worked
    if (ioctl(loop_fd, LOOP_GET_STATUS64, &li) < 0) {
        fprintf(stderr, "ERROR: Unable to get loop status for %s (%s)", filename,
             strerror(errno));
       return -1;
    }

    close(loop_fd);
    close(image_fd);

    return 0;
}


/* Disconnect and delete the loopback device */
int rmloop(const char *loopDevice) {
    int device_fd;
    struct loop_info64 li;

    // open the loopback device
    if ((device_fd = open(loopDevice, O_RDONLY)) < 0) {
        fprintf(stderr, "ERROR: Failed to open loop (%d)", errno);
        return -1;
    }

    // get the status of the device and store in a loop_info struct
    if (ioctl(device_fd, LOOP_GET_STATUS64, &li) < 0) {
        fprintf(stderr, "ERROR: Unable to get loop status for %s (%s)", loopDevice,
             strerror(errno));
       return -1;
    }

    // disconnect the loopback device from the associated image file
    if (ioctl(device_fd, LOOP_CLR_FD, 0) < 0) {
        fprintf(stderr, "ERROR: Failed to destroy loop (%d)", errno);
        close(device_fd);
        return -1;
    }

    close(device_fd);

    // delete the loopback block device file
    if (remove(loopDevice) < 0) {
        fprintf(stderr, "ERROR: Failed to delete node file %s (%d)", loopDevice, errno);
        return -1;
    }
    return 0;
}

// main function
int main(int argc, char *argv[])
{
  unsigned char salt[SALT_LEN];      		    // salt read from disk image
  unsigned char result[IV_LEN_BYTES + KEY_LEN_BYTES]; // result of PBKDF2 (intermediate key)
  unsigned char encrypted_key[KEY_LEN_BYTES];	// key read from disk
  unsigned char decrypted_key[KEY_LEN_BYTES];	// derived master key
  unsigned char dm[MAXPATHLEN];                 // path to device mapper file
  char passwd[1024], buffer[1024];              // some buffers
  struct crypt_mnt_ftr crypt_ftr;               // the filesystem footer structure


  // print usage info
  if ( argc != 5 && argc != 3) {
    fprintf(stderr, "\n");
    fprintf(stderr, "%s\n", argv[0]);
    fprintf(stderr, "mounts an encrypted android filesystem image or block device\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "usage: \n");
    fprintf(stderr, "\n");
    fprintf(stderr, "sudo %s  <passwd>  <path_to_image>  <mount_point>  <pwd|key>\n", argv[0]);
    fprintf(stderr, "\n");
    fprintf(stderr, "where <pwd|key> chooses a password or key (keys are already PBKDF2 decrypted)\n"); 
    fprintf(stderr, "and <passwd> is the associated string\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "sudo %s  -d  <mount_point>\n", argv[0]);
    fprintf(stderr, "\n");
    fprintf(stderr, "to unmount / cleanup a mounted image\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "\n");
    exit(1);
  }

// mount an image file
if (argc == 5) {

  // read crypt footer from disk image and extract salt & encrypted master key
  if (read_crypt_ftr(argv[2], &crypt_ftr, encrypted_key, salt) < 0) {
    fprintf(stderr, "MAIN ERROR: Cannot read crypt footer from %s\n", argv[2]);
    exit(2);
  }

  // Print footer info to stderr
  fprintf(stderr, "\nFooter Magic Number: \t%08x\n", crypt_ftr.magic);
  bin_to_hex(encrypted_key, KEY_LEN_BYTES, buffer);
  fprintf(stderr, "Master Key (encrypted): %s\n", buffer);
  bin_to_hex(salt, SALT_LEN, buffer);
  fprintf(stderr, "Salt: \t\t\t%s\n", buffer);
  fprintf(stderr, "Cipher: \t\t%s\n", crypt_ftr.crypto_type_name);
  fprintf(stderr, "FS Size (in sectors): \t%d\n", (int)crypt_ftr.fs_size);





/* *******************************************************
 *   XXX FIX ME
     XXX ADD USE OF KEY WITHOUT DOING PBKDF2 OR MASTER KEY DECRYPTION...
 * ****************************************************** */
   if ((strcmp(argv[4], "pwd") == 0)) {
      sprintf(passwd, "%d", atoi(argv[1]));
   } else if (strcmp(argv[4], "key") == 0) {
      sprintf(passwd, "%d", atoi(argv[1]));
      fprintf(stderr, "Sorry %s functionality is not yet implemented, please use pwd\n", argv[4]);
      exit(2);
   } else {
      fprintf(stderr, "MAIN ERROR: Unknown unlock type %s\n", argv[4]);
      exit(2);
   }

   fprintf(stderr, "Using: \t\t\t%s %s\n", argv[4], passwd);

/* ********************************************************* */


   // Perform PBKDF2
   pbkdf2(passwd, salt, result);

   // Decrypt master key
   if (decrypt_master_key(result, encrypted_key, decrypted_key) < 0) {
       fprintf(stderr, "MAIN ERROR: Cannot decrypt master key\n");
       exit(2);
   }

  bin_to_hex(decrypted_key, KEY_LEN_BYTES, buffer);
  fprintf(stderr, "Master Key (decrypted): %s\n", buffer);

   // create loopback block device from image (Encrypted)
   if (mkloop(argv[2]) < 0) {
       fprintf(stderr, "MAIN ERROR: Cannot create loopback device %s from file %s\n", "/dev/loop254", argv[2]);
       exit(2);
   }

   // Create dm-crypt mapped block device (Decrypted)
   if (mkdevmap(&crypt_ftr, decrypted_key, "/dev/loop254", dm, "droid-fs") < 0) {
       fprintf(stderr, "MAIN ERROR: Cannot create dm-crypt device %s from loopback %s\n", dm, "/dev/loop254");
       exit(2);
   }

   // Mount the devmapper device (try EXT4 and FAT)
   mkdir(argv[2], 0755);
   if ( mount(dm, argv[3], "ext4", MS_RDONLY|MS_NODIRATIME|MS_NOATIME, "") ) {
     fprintf(stderr, "WARNING: Cannot mount %s as EXT4 on %s\n",dm, argv[3]);
        if ( mount(dm, argv[3], "vfat", MS_RDONLY|MS_NODIRATIME|MS_NOATIME, "") ) {
          fprintf(stderr, "MAIN ERROR: Cannot mount %s as FAT on %s either\n",dm, argv[3]);
          exit(2);
        }
   }
          // If we got here, everything worked!
          fprintf(stderr, "\n\nSUCCESS:\n\t\tImage %s \n\t\ton Loopback %s \n\t\tMapped to %s \n\t\tMounted on %s\n\n", argv[2], "/dev/loop254", dm, argv[3]);


// -d cleanup command invoked
} else if (argc == 3) {	
    umount(argv[2]);
    rmdevmap("droid-fs");
    rmloop("/dev/loop254");
}

  return(0);
}
