/*
 * Copyright (C) 2010 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 * 
 * This file contains adapted source code from the AOSP project
 * modified by Adam Skillen to provide the password recovery functionality.
 *   
 * Copyright (C) 2013 Adam Skillen
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Adam Skillen
 * 4. Neither the name of Concordia University, Mobiflage, nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *  
 */

/*
 * FILE:		mkdict.c
 * 
 * AUTHOR:		Adam Skillen
 * DATE:		13-Feb-12
 *
 * DESCRIPTION:
 * This file creates a dictionary of potential decryption keys for Android 4.0 (ICS)
 * encrypted file systems.  It takes as input an encrypted disk image and password 
 * options, and outputs the derived master keys that can be used to decrypt the file system.
 * The password used to unlock the device may be a PIN, a pattern, or an ASCII passphrase.
 *
 *  USAGE: mkdict  <pin | pwd>  <max_length | path_to_dictionary>  <path_to_image> 
 *  BUILD: gcc -o mkdict mkdict.c -lcrypto -lm
 * 
 * NOTICE: 
 * Some of this code is derived from the Android Open Source Project (AOSP).
 * Please see the [aosp_src]/system/vold directory for the original files.
 * The AOSP license information can be found above this text.
*/

#define _LARGEFILE64_SOURCE 1

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <linux/types.h>
#include <linux/dm-ioctl.h>
#include <linux/loop.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <fcntl.h>
#include <unistd.h>
#include <openssl/evp.h>
#include <openssl/hmac.h>


#define KEY_LEN_BYTES 16
#define IV_LEN_BYTES 16
#define CRYPT_FOOTER_OFFSET 0x4000
#define SALT_LEN 16
#define KEY_TO_SALT_PADDING 32
#define CRYPT_MNT_MAGIC 0xD0B5B1C4
#define DM_CRYPT_BUF_SIZE 4096
#define MAXPATHLEN 64
#define DEVMAPPER_BUFFER_SIZE 4096
#define HASH_COUNT 2000

#define __le32 unsigned int
#define __le16 unsigned short int 

// crypto footer structure format
struct crypt_mnt_ftr {
  __le32 magic;		/* crypto footer magic number */
  __le16 major_version; 
  __le16 minor_version;
  __le32 ftr_size; 	/* footer size in bytes (not including key) */
  __le32 flags;		/* flags for Android OS */
  __le32 keysize;	/* keylength in bytes */
  __le32 spare1;	/* ignored */
  __le64 fs_size;	/* Size of the encrypted fs, in 512 byte sectors */
  __le32 failed_decrypt_count;
  unsigned char crypto_type_name[64];	/* cipher, mode, IV method, etc. */
};


/* Convert a binary string of specified length into an ascii hex string equivalent */
void bin_to_hex(unsigned char *bin_buf, unsigned int length,
                              char *hex_ascii)
{
  unsigned int i, a;
  unsigned char nibble;

  for (i=0, a=0; i<length; i++, a+=2) {
    /* For each byte, write out two ascii hex digits */
    nibble = (bin_buf[i] >> 4) & 0xf;
    hex_ascii[a] = nibble + (nibble > 9 ? 0x37 : 0x30);

    nibble = bin_buf[i] & 0xf;
    hex_ascii[a+1] = nibble + (nibble > 9 ? 0x37 : 0x30);
  }
  hex_ascii[a] = '\0';
}


/* Turn the password into a key and IV that can decrypt the master key 
   Using OpenSSL PBKDF2 library function given a password and salt and
   iterating 2000 times */
static void pbkdf2(char *passwd, unsigned char *salt, unsigned char *ikey)
{
    PKCS5_PBKDF2_HMAC_SHA1(passwd, strlen(passwd), salt, SALT_LEN,
                           HASH_COUNT, KEY_LEN_BYTES+IV_LEN_BYTES, ikey);
}

/* Read the footer from a disk image and store it in a footer struct */
static int read_crypt_ftr(char *disk_image, struct crypt_mnt_ftr *crypt_ftr,
                                  unsigned char *key, unsigned char *salt)
{
  int fd;
  unsigned int cnt;
  off64_t off;
  int rc = -1;
  struct stat64 statbuf;

  // open the disk image for reading
  if ( (fd = open(disk_image, O_RDONLY)) < 0) {
    fprintf(stderr, "ERROR: Cannot open disk image %s\n", disk_image);
    return -1;
  }

  // get the stat info for the image
  if (fstat64(fd, &statbuf) < 0) {
    fprintf(stderr, "ERROR: Cannot stat(2) disk image %s\n", disk_image);
    goto errout;
  }

  off = ((off64_t)statbuf.st_size) - CRYPT_FOOTER_OFFSET;

  // seek to the footer area in the image
  if (lseek64(fd, off, SEEK_SET) == -1) {
    fprintf(stderr, "ERROR: Cannot seek to footer\n");
    goto errout;
  }

  // read the footer and store it in a struct
  if ( (cnt = read(fd, crypt_ftr, sizeof(struct crypt_mnt_ftr))) != sizeof(struct crypt_mnt_ftr)) {
    fprintf(stderr, "ERROR: Cannot read disk image %s\n", disk_image);
    goto errout;
  }

  // sanity check: ensure proper magic number in footer
  if (crypt_ftr->magic != CRYPT_MNT_MAGIC) {
    fprintf(stderr, "ERROR: Bad magic number in crypt footer: \n Wanted %08x \n Found %08x\n", CRYPT_MNT_MAGIC, crypt_ftr->magic);
    goto errout;
  }

  // if the footer has custom length, skip to end of the footer to read the key and salt
  if (crypt_ftr->ftr_size > sizeof(struct crypt_mnt_ftr)) {
    if (lseek(fd, crypt_ftr->ftr_size - sizeof(struct crypt_mnt_ftr),  SEEK_CUR) == -1) {
      fprintf(stderr, "ERROR: Cannot seek to key\n");
      goto errout;
    }
  }

  // sanity check: ensure proper key length is used (causes a warning, does not halt)
  if (crypt_ftr->keysize != KEY_LEN_BYTES) {
    fprintf(stderr, "WARNING: Keysize of %d bits may not be supported for image %s\n",
          crypt_ftr->keysize * 8, disk_image);
  }

  // read the key stored after the footer
  if ( (cnt = read(fd, key, crypt_ftr->keysize)) != crypt_ftr->keysize) {
    fprintf(stderr, "ERROR: Cannot read key from image %s\n", disk_image);
    goto errout;
  }

  // seek to salt stored after the key
  if (lseek64(fd, KEY_TO_SALT_PADDING, SEEK_CUR) == -1) {
    fprintf(stderr, "ERROR: Cannot seek to salt\n");
    goto errout;
  }

  // read the salt (used for PBKDF2)
  if ( (cnt = read(fd, salt, SALT_LEN)) != SALT_LEN) {
    fprintf(stderr, "ERROR: Cannot read salt from image %s\n", disk_image);
    goto errout;
  }
  rc = 0;

errout:
  close(fd);
  return rc;
}

/* using the PBKDF2 derived key and salt, decrypt the master key */
static int decrypt_master_key(unsigned char *ikey,
                              unsigned char *encrypted_master_key,
                              unsigned char *decrypted_master_key)
{
  EVP_CIPHER_CTX d_ctx;
  int decrypted_len, final_len;

  /* Initialize the decryption engine */
  if (! EVP_DecryptInit(&d_ctx, EVP_aes_128_cbc(), ikey, ikey+KEY_LEN_BYTES)) {
    return -1;
  }
  EVP_CIPHER_CTX_set_padding(&d_ctx, 0); /* Turn off padding as our data is block aligned */

  /* Decrypt the master key */
  if (! EVP_DecryptUpdate(&d_ctx, decrypted_master_key, &decrypted_len,
                            encrypted_master_key, KEY_LEN_BYTES)) {
    return -1;
  }
  if (! EVP_DecryptFinal(&d_ctx, decrypted_master_key + decrypted_len, &final_len)) {
    return -1;
  }

  // sanity check: is the decrypted string the proper length
  if (decrypted_len + final_len != KEY_LEN_BYTES) {
    return -1;
  } else {
    return 0;
  }
}

/* int wrapper for exponentiation */
int power(int base, int exp) {
	return (int) pow((double) base, (double) exp);
}


/* main function */
int main(int argc, char *argv[])
{
  unsigned char salt[SALT_LEN];      		          // salt read from disk image
  unsigned char result[IV_LEN_BYTES + KEY_LEN_BYTES]; // result of PBKDF2 (intermediate key)
  unsigned char encrypted_key[KEY_LEN_BYTES];	      // key read from disk
  unsigned char decrypted_key[KEY_LEN_BYTES];	      // derived master key
  unsigned int pin, len, max;
  unsigned char dm[MAXPATHLEN];                       // path to device mapper file
  char passwd[1024], buffer[1024], pwd[128];          // string buffers
  int fd;
  char *passwd_ptr;
  struct crypt_mnt_ftr crypt_ftr;                    // the footer structure


  // print usage info
  if ( argc != 4) {
    fprintf(stderr, "\n");
    fprintf(stderr, "%s\n", argv[0]);
    fprintf(stderr, "creates a dictionary to crack an android filesystem image or block device\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "usage: \n");
    fprintf(stderr, "\n");
    fprintf(stderr, "usage: %s  <pin | pwd>  <max_length | path_to_dictionary>  <path_to_image> \n", argv[0]);
    fprintf(stderr, "\n");
    fprintf(stderr, "where <pin | pwd> chooses between PIN codes or a password dictionary file\n"); 
    fprintf(stderr, "and <max_len> is the maximum length for unlock code if pin is used\n");
    fprintf(stderr, "and <path_to_dictionary> is the dictionary file if pwd is used\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "You will want to redirect stdout to a file for use with the dc tool\n");
    fprintf(stderr, "\n");
    exit(1);
  }

  // read crypt footer from disk image and extract salt & encrypted master key
  if (read_crypt_ftr(argv[3], &crypt_ftr, encrypted_key, salt) < 0) {
    fprintf(stderr, "MAIN ERROR: Cannot read crypt footer from %s\n", argv[3]);
    exit(2);
  }

  // Print footer info to stderr
  fprintf(stderr, "\nFooter Magic Number: \t%08x\n", crypt_ftr.magic);
  bin_to_hex(encrypted_key, KEY_LEN_BYTES, buffer);
  fprintf(stderr, "Master Key (encrypted): %s\n", buffer);
  bin_to_hex(salt, SALT_LEN, buffer);
  fprintf(stderr, "Salt: \t\t\t%s\n", buffer);
  fprintf(stderr, "Cipher: \t\t%s\n", crypt_ftr.crypto_type_name);
  fprintf(stderr, "FS Size (in sectors): \t%d\n", (int)crypt_ftr.fs_size);


   // using PIN
   if (strcmp(argv[1], "pin") == 0) {
     len = atoi(argv[2]) < 4 ? 4 : atoi(argv[2]);

    // this outer loop makes pins of all lengths up to len
    // e.g., if len==6 then 6, 5 AND 4 digit pins will be generated
    do {

     // loop over PIN codes
     max = power(10, len);
     for (pin = 0; pin < max; pin++) {

       // Perform PBKDF2
       sprintf(passwd, "%0*d", len, pin);
       pbkdf2(passwd, salt, result);

       // Decrypt master key
       if (decrypt_master_key(result, encrypted_key, decrypted_key) < 0) {
           fprintf(stderr, "MAIN ERROR: Cannot decrypt master key\n");
           exit(2);
       }

       // Print PIN/Decrypted Master Key to stdout
       bin_to_hex(decrypted_key, KEY_LEN_BYTES, buffer);
       printf("%8s/%s\n", passwd, buffer);
       
     }
    } while (len-- > 4);

   // using password dictionary
   } else if (strcmp(argv[1], "pwd") == 0) {


      // load dictionary file
      if ((fd = open(argv[2], O_RDONLY)) < 0) {
        fprintf(stderr, "ERROR: Cannot open password file %s\n", argv[2]);
        exit(2);
      }

      // iterate over entries in file
      passwd_ptr = pwd - 1;
      while (read(fd, ++passwd_ptr, sizeof(char))) {

          if (*passwd_ptr == '\n') {
              *passwd_ptr = '\0';

            // Perform PBKDF2
            pbkdf2(pwd, salt, result);

            // Decrypt master key
            if (decrypt_master_key(result, encrypted_key, decrypted_key) < 0) {
                fprintf(stderr, "MAIN ERROR: Cannot decrypt master key\n");
                exit(2);
            }

            // Print "[Password]/[Decrypted Master Key]" to stdout
            bin_to_hex(decrypted_key, KEY_LEN_BYTES, buffer);
            printf("%12s/%s\n", pwd, buffer);

            passwd_ptr = pwd - 1;
            }
         }
   } else {
      fprintf(stderr, "MAIN ERROR: Unknown unlock type %s\n", argv[1]);
      exit(2);
   }


  return(0);
}
